/*
 * Copyright (C) 2012 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package android.content.pm;

import android.util.Slog;
import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.security.DigestInputStream;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.util.Arrays;
import libcore.io.IoUtils;

/**
 * Represents the manifest digest for a package. This is suitable for comparison
 * of two packages to know whether the manifests are identical.
 *
 * @hide
 */
public class ManifestDigest {
	private static final String TAG = "ManifestDigest";

	/**
	 * The digest of the manifest in our preferred order.
	 */
	private final byte[] mDigest;

	/**
	 * What we print out first when toString() is called.
	 */
	private static final String TO_STRING_PREFIX = "ManifestDigest {mDigest=";

	/**
	 * Digest algorithm to use.
	 */
	private static final String DIGEST_ALGORITHM = "SHA-256";

	ManifestDigest(byte[] digest) {
		mDigest = digest;
	}

	static ManifestDigest fromInputStream(InputStream fileIs) {
		if (fileIs == null) {
			return null;
		}

		final MessageDigest md;
		try {
			md = MessageDigest.getInstance(DIGEST_ALGORITHM);
		} catch (NoSuchAlgorithmException e) {
			throw new RuntimeException(DIGEST_ALGORITHM + " must be available", e);
		}

		final DigestInputStream dis = new DigestInputStream(new BufferedInputStream(fileIs), md);
		try {
			byte[] readBuffer = new byte[8192];
			while (dis.read(readBuffer, 0, readBuffer.length) != -1) {
				// not using
			}
		} catch (IOException e) {
			Slog.w(TAG, "Could not read manifest");
			return null;
		} finally {
			IoUtils.closeQuietly(dis);
		}

		final byte[] digest = md.digest();
		return new ManifestDigest(digest);
	}

	public int describeContents() {
		return 0;
	}

	@Override
	public boolean equals(Object o) {
		if (!(o instanceof ManifestDigest)) {
			return false;
		}

		final ManifestDigest other = (ManifestDigest)o;

		return this == other || Arrays.equals(mDigest, other.mDigest);
	}

	@Override
	public int hashCode() {
		return Arrays.hashCode(mDigest);
	}

	@Override
	public String toString() {
		final StringBuilder sb = new StringBuilder(TO_STRING_PREFIX.length() + (mDigest.length * 3) + 1);

		sb.append(TO_STRING_PREFIX);

		final int N = mDigest.length;
		for (int i = 0; i < N; i++) {
			final byte b = mDigest[i];
			sb.append(String.format("%02x", b));
			sb.append(',');
		}
		sb.append('}');

		return sb.toString();
	}
}
